import json
import logging
from typing import IO, Any, Dict, Iterator, List, Union

LOG = logging.getLogger(__name__)


class JSONReader:
    """
    A JSON reader that optionally iterates over a json list
    """

    def __init__(
        self,
        io_str: IO[str],
        required_properties: List[str] = None,
        json_path: List[Union[str, int]] = None,
        name: str = 'json file',
    ):
        """
        :param io_str: Any IO stream that yields a string
                       See https://docs.python.org/3/library/io.html#io.IOBase
        :param required_properties: required top level properties
        :param iterate_over: todo
        :param name: todo
        """
        self.io_str = io_str
        self.required_properties = required_properties
        self.json_path = json_path
        self.name = name

        if self.json_path:
            self.json_obj = json.load(self.io_str)
            for path in self.json_path:
                self.json_obj = self.json_obj[path]
        else:
            self.json_obj = json.load(self.io_str)

        if isinstance(self.json_obj, list):
            self._len = len(self.json_obj)
            self._line_num = 0
        else:
            self.json_obj = [self.json_obj]
            self._len = 0
            self._line_num = 0

    def __iter__(self) -> Iterator:
        return self

    def __next__(self) -> Dict[str, Any]:

        if self._line_num + 1 > self._len:
            LOG.info(f"Finished processing {self.name}")
            raise StopIteration

        next_obj = self.json_obj[self._line_num]

        self._line_num = self._line_num + 1

        if self.required_properties:
            if not set(next_obj.keys()) >= set(self.required_properties):
                # TODO - have koza runner handle this exception
                # based on some configuration? similar to
                # on_map_error
                raise ValueError(
                    f"Configured properties missing in source file "
                    f"{set(self.required_properties) - set(next_obj.keys())}"
                )

            # If we want to subset
            # next_obj = {key: next_obj[key] for key in next_obj.keys() if key in self.required_properties}

        return next_obj



"""
End to end test of String load from examples/string

"""

import os

import pytest

from koza.koza_runner import transform_source
from koza.model.config.source_config import OutputFormat


@pytest.mark.parametrize(
    "ingest, output_names, output_format",
    [
        (
            "string",
            [
                "stringdb.protein-links-detailed",
            ],
            OutputFormat.tsv,
        ),
        (
            "string",
            [
                "stringdb.protein-links-detailed",
            ],
            OutputFormat.jsonl,
        ),
        (
            "string-declarative",
            [
                "stringdb.protein-links-detailed",
            ],
            OutputFormat.tsv,
        ),
        (
            "string-declarative",
            [
                "stringdb.protein-links-detailed",
            ],
            OutputFormat.jsonl,
        ),
    ],
)
def test_examples(ingest, output_names, output_format):

    source = f"examples/{ingest}/metadata.yaml"
    output_suffix = str(output_format).split('.')[1]
    output_dir = f"./test-output/{ingest}-{output_suffix}"

    output_files = [
        output_file
        for file in output_names
        for output_file in [
            f"{output_dir}/{file}_nodes.{output_suffix}",
            f"{output_dir}/{file}_edges.{output_suffix}",
        ]
    ]

    transform_source(source, output_dir, output_format, "examples/translation_table.yaml", None)

    for file in output_files:
        assert os.path.exists(file)
        assert os.path.getsize(file) > 0

    # TODO: at some point, these assertions could get more rigorous, but knowing if we have errors/exceptions is a start
    # TODO: kgx validation could also be added back in, especially if something programatic is done with the output



import importlib
import logging
import os
import sys
from pathlib import Path
from typing import Dict, Iterable

import yaml
from pydantic.error_wrappers import ValidationError

from koza.exceptions import MapItemException
from koza.io.writer.jsonl_writer import JSONLWriter
from koza.io.writer.tsv_writer import TSVWriter
from koza.io.writer.writer import KozaWriter
from koza.model.config.source_config import MapFileConfig, OutputFormat
from koza.model.curie_cleaner import CurieCleaner
from koza.model.map_dict import MapDict
from koza.model.source import Source, SourceFile

LOG = logging.getLogger(__name__)


class KozaApp:
    """
    Class that holds all configuration information for Koza

    Note that this is intended to be a singleton
    that is instantiated in biolink_runner and that
    object either imported in other modules, or
    passed in via a function

    Note that this borders on some anti-patterns
    - singleton as global
    - god object

    So this particular code should be re-evaluated post prototype

    Hoping that making all attributes read-only offsets some downsides
    of this approach (multi threading would be fine)
    """

    def __init__(
        self,
        source: Source,
        output_dir: str = './output',
        output_format: OutputFormat = OutputFormat('jsonl'),
    ):
        self.source = source
        self.output_dir = output_dir
        self.output_format = output_format
        self.file_registry: Dict[str, SourceFile] = {}
        self.map_registry: Dict[str, SourceFile] = {}
        self.writer_registry: Dict[str, KozaWriter] = {}
        self.map_cache: Dict[str, Dict] = {}
        self.curie_cleaner: CurieCleaner = CurieCleaner()

        logging.getLogger(__name__)

        for source_file in source.source_files:

            if not source_file.config.transform_code:
                # look for it alongside the source conf as a .py file
                source_file.config.transform_code = (
                    str(source_file.config.path.parent / source_file.config.path.stem) + '.py'
                )

            if source_file.config.depends_on is not None:
                for map_file in source_file.config.depends_on:
                    if map_file_config.name not in self.map_cache:
                        with open(map_file, 'r') as map_file_fh:
                            map_file_config = MapFileConfig(**yaml.safe_load(map_file_fh))
                            map_file_config.transform_code = (
                                str(Path(map_file).parent / Path(map_file).stem) + '.py'
                            )
                        self.map_registry[map_file_config.name] = SourceFile(map_file_config)

            self.file_registry[source_file.config.name] = source_file

            output_name = f"{source.name}.{source_file.config.name}"
            self.writer_registry[source_file.config.name] = self.get_writer(
                output_name, source_file.config.node_properties, source_file.config.edge_properties
            )

    def get_writer(self, name, node_properties, edge_properties):
        if self.output_format == OutputFormat.tsv:
            return TSVWriter(self.output_dir, name, node_properties, edge_properties)

        elif self.output_format == OutputFormat.jsonl:
            return JSONLWriter(self.output_dir, name)

    def get_map(self, map_name: str):
        pass

    def process_sources(self):
        """
        :return:
        """
        import sys

        for map_file in self.map_registry.values():
            if map_file.config.name not in self.map_cache:
                self.load_map(map_file.config)

        for source_file in self.file_registry.values():
            parent_path = Path(source_file.config.transform_code).parent
            transform_code = Path(source_file.config.transform_code).stem
            sys.path.append(str(parent_path))
            is_first = True
            transform_module = None

            if source_file.config.transform_mode == 'flat':
                while True:
                    try:
                        if is_first:
                            transform_module = importlib.import_module(transform_code)
                            is_first = False
                        else:
                            importlib.reload(transform_module)
                    except MapItemException as mie:
                        LOG.warning(f"{str(mie)} not found in map")
                    except ValidationError as ve:
                        LOG.error(f"Validation error while processing: {source_file.last_row}")
                        raise ve
                    except StopIteration:
                        break
            elif source_file.config.transform_mode == 'loop':
                if transform_code not in sys.modules.keys():
                    importlib.import_module(transform_code)
                else:
                    importlib.reload(importlib.import_module(transform_code))
            else:
                raise NotImplementedError

            # close the writer when the source is done processing
            self.writer_registry[source_file.config.name].finalize()

            # remove directory from sys.path to prevent name clashes
            sys.path.remove(str(parent_path))

    def load_map(self, map_file_config: MapFileConfig):
        source_file = SourceFile(map_file_config)

        map = MapDict()

        self.map_cache[map_file_config.name] = map

        if os.path.exists(map_file_config.transform_code):
            parent_path = Path(map_file_config.transform_code).parent
            transform_code = Path(map_file_config.transform_code).stem
            sys.path.append(str(parent_path))
            is_first = True
            transform_module = None

            while True:
                try:
                    if is_first:
                        transform_module = importlib.import_module(transform_code)
                        is_first = False
                    else:
                        importlib.reload(transform_module)
                except StopIteration:
                    break
        else:
            key_column = map_file_config.key
            value_columns = map_file_config.values
            for row in source_file:
                map[row[key_column]] = {
                    key: value for key, value in row.items() if key in value_columns
                }

    def write(self, source_name, entities: Iterable):
        self.writer_registry[source_name].write(entities)



"""
Module for managing koza runs
"""

import logging
from pathlib import Path
from typing import List, Optional

import yaml

from koza.app import KozaApp
from koza.io.reader.csv_reader import CSVReader
from koza.io.reader.json_reader import JSONReader
from koza.io.reader.jsonl_reader import JSONLReader
from koza.io.utils import open_resource
from koza.model.config.source_config import CompressionType, FormatType, OutputFormat, SourceConfig
from koza.model.source import Source
from koza.model.translation_table import TranslationTable

LOG = logging.getLogger(__name__)

KOZA_APP = None


def set_koza(koza: KozaApp):
    global KOZA_APP
    KOZA_APP = koza


def set_koza_app(
    source: Source,
    output_dir: str = './output',
    output_format: OutputFormat = OutputFormat('jsonl'),
) -> KozaApp:
    """
    Setter for singleton koza app object
    """
    global KOZA_APP

    KOZA_APP = KozaApp(source, output_dir, output_format)
    return KOZA_APP


def get_koza_app() -> Optional[KozaApp]:
    """
    Getter for singleton koza app object
    """
    return KOZA_APP


def validate_file(
    file: str,
    format: FormatType = FormatType.csv,
    delimiter: str = ',',
    header_delimiter: str = None,
    compression: CompressionType = None,
    skip_lines: int = 0,
    skip_blank_lines: bool = True,
):
    """
    Runs a file through one of the Koza readers
    For csv files will return header and row length

    For json and jsonl just validates them
    """

    with open_resource(file, compression) as resource_io:

        if format == FormatType.csv:
            reader = CSVReader(
                resource_io,
                delimiter=delimiter,
                header_delimiter=header_delimiter,
                skip_lines=skip_lines,
                skip_blank_lines=skip_blank_lines,
            )
        elif format == FormatType.jsonl:
            reader = JSONLReader(resource_io)
        elif format == FormatType.json:
            reader = JSONReader(resource_io)
        else:
            raise ValueError

        for _ in reader:
            pass


def validate_source_files(source: str, source_files: List[str]) -> List[str]:
    """
    Resolve a source string to its source metadata file

    :param source:
    :param source_files:
    :return:
    """
    source_dir = Path(source).parent
    source_paths = []
    for src_file in source_files:
        if not (Path(src_file).exists() or Path(src_file).is_file()):
            # check if it's in source parent
            updated_src_file = source_dir / src_file
            if not (Path(updated_src_file).exists() or Path(updated_src_file).is_file()):
                raise ValueError(f"Source file doesn't exist: {src_file}")
            else:
                source_paths.append(updated_src_file)
        else:
            source_paths.append(src_file)
    return source_paths


def get_translation_table(global_table: str, local_table: str) -> TranslationTable:
    # Look for translation table files

    global_tt = {}
    local_tt = {}

    if not global_table:
        if local_table:
            raise ValueError("Local table without a global table not allowed")
        else:
            LOG.info("No global table used for transform")
    else:

        with open(global_table, 'r') as global_tt_fh:
            global_tt = yaml.safe_load(global_tt_fh)

        if local_table:
            with open(local_table, 'r') as local_tt_fh:
                local_tt = yaml.safe_load(local_tt_fh)
        else:
            LOG.info("No local table used for transform")

    return TranslationTable(global_tt, local_tt)


def transform_source(
    source: str,
    output_dir: str,
    output_format: OutputFormat,
    global_table: str,
    local_table: str,
):

    translation_table = get_translation_table(global_table, local_table)

    with open(source, 'r') as source_fh:
        source_config = SourceConfig(**yaml.safe_load(source_fh))
        if not source_config.name:
            source_config.name = Path(source).stem

        source = Source(
            source_files=validate_source_files(source, source_config.source_files),
            name=source_config.name,
            dataset_description=source_config.dataset_description,
            translation_table=translation_table,
        )

        koza_app = set_koza_app(source, output_dir, output_format)
        koza_app.process_sources()



import types
from typing import Iterable

import pytest

from koza.app import KozaApp
from koza.koza_runner import set_koza
from koza.model.config.source_config import PrimaryFileConfig
from koza.model.source import Source, SourceFile


@pytest.fixture(scope="package")
def mock_koza():

    # This should be extracted out but for quick prototyping
    def _mock_write(self, source_name, *entities):
        self._entities = list(*entities)

    def _make_mock_koza_app(
        name: str,
        data: Iterable,
        transform_code: str,
        map_cache=None,
        filters=None,
        translation_table=None,
    ):
        mock_source_file_config = PrimaryFileConfig(
            name=name,
            files=[],
            transform_code=transform_code,
        )
        mock_source_file = SourceFile(mock_source_file_config)
        mock_source_file._reader = data

        mock_source = Source(source_files=[mock_source_file])

        koza = KozaApp(mock_source)
        # TODO filter mocks
        koza.source.translation_table = translation_table
        koza.map_cache = map_cache
        koza.write = types.MethodType(_mock_write, koza)

        return koza

    def _transform(
        name: str,
        data: Iterable,
        transform_code: str,
        map_cache=None,
        filters=None,
        translation_table=None,
    ):
        koza_app = _make_mock_koza_app(
            name,
            data,
            transform_code,
            map_cache=map_cache,
            filters=filters,
            translation_table=translation_table,
        )
        set_koza(koza_app)
        koza_app.process_sources()
        return koza_app._entities

    return _transform



import logging
from csv import reader
from typing import IO, Any, Dict, Iterator

from koza.model.config.source_config import FieldType

LOG = logging.getLogger(__name__)


FIELDTYPE_CLASS = {
    FieldType.str: str,
    FieldType.int: int,
    FieldType.float: float,
    # FieldType.Proportion: Proportion,
}


class CSVReader:
    """
    A CSV reader modelled after csv.DictReader

    https://docs.python.org/3/library/csv.html#csv.DictReader

    The differences are:
      - Checking of field names against the header, returning
        a warning if extra fields exist and an ValueError
        if a field is missing

      - Handle rows that are shorter/longer than expected
        currently handled by raising an exception

      - Support a type map dictionary, in which fieldnames
        can be mapped to types, and the CSVReader will attempt
        to coerce them from their str representation, eg int('42')

      - Potentially will add support a multivalued field DSL, eg
        List[str][';'] would convert a semicolon delimited multivalued
        field to a list of strings

    TODO handle cases when delimiter is >1 character
    """

    def __init__(
        self,
        io_str: IO[str],
        field_type_map: Dict[str, FieldType] = None,
        delimiter: str = ",",
        has_header: bool = True,
        header_delimiter: str = None,
        dialect: str = "excel",
        skip_lines: int = 0,
        skip_blank_lines: bool = True,
        name: str = "csv file",
        *args,
        **kwargs,
    ):
        """
        :param io_str: Any IO stream that yields a string
                       See https://docs.python.org/3/library/io.html#io.IOBase
        :param name: filename or alias
        :param field_type_map: A dictionary of field names and their type (using the FieldType enum)
        :param delimiter: Field delimiter (eg. '\t' ',' ' ')
        :param has_header: true if the file has a header, default=True
        :param header_delimiter: delimiter for the header row, default = self.delimiter
        :param dialect: csv dialect, default=excel
        :param args: additional args to pass to csv.reader
        :param kwargs: additional kwargs to pass to csv.reader
        """
        self.io_str = io_str
        self.field_type_map = field_type_map
        self.dialect = dialect
        self.has_header = has_header
        self.header_delimiter = header_delimiter if header_delimiter else delimiter
        self.skip_lines = skip_lines
        self.skip_blank_lines = skip_blank_lines
        self.name = name

        self.line_num = 0
        self.fieldnames = []

        if delimiter == '\\s':
            delimiter = ' '

        kwargs['dialect'] = dialect
        kwargs['delimiter'] = delimiter
        self.reader = reader(io_str, *args, **kwargs)

    def __iter__(self) -> Iterator:
        return self

    def __next__(self) -> Dict[str, Any]:

        while self.line_num < self.skip_lines:
            next(self.reader)
            self.line_num = self.reader.line_num

        if self.line_num == self.skip_lines:

            if not self.has_header and not self.field_type_map:
                raise ValueError(
                    f"there is no header and columns have not been supplied\n"
                    f"configure the 'columns' property in the source yaml"
                )

            if self.has_header:
                fieldnames = next(
                    reader(
                        self.io_str, **{'delimiter': self.header_delimiter, 'dialect': self.dialect}
                    )
                )
                # todo: maybe comment character should be specified?
                fieldnames[0] = fieldnames[0].lstrip('#')
                fieldnames[0] = fieldnames[0].lstrip('!!')
                fieldnames = [f.strip() for f in fieldnames]
            else:
                fieldnames = list(self.field_type_map.keys())

            self.fieldnames = fieldnames

            if self.field_type_map:

                configured_fields = list(self.field_type_map.keys())

                if set(configured_fields) > set(fieldnames):
                    raise ValueError(
                        f"Configured columns missing in source file {self.name}\n"
                        f"{set(configured_fields) - set(fieldnames)}"
                    )

                if set(fieldnames) > set(configured_fields):
                    LOG.warning(
                        f"Additional column(s) in source file {self.name}\n"
                        f"{set(fieldnames) - set(configured_fields)}\n"
                        f"Checking if new column(s) inserted at end of the row"
                    )
                    # add to type map
                    for new_fields in set(fieldnames) - set(configured_fields):
                        self.field_type_map[new_fields] = FieldType.str

                # Check if the additional columns are appended
                # not sure if this would useful or just noise
                if fieldnames[: len(configured_fields)] != configured_fields:
                    LOG.warning(
                        f"Additional columns located within configured fields\n"
                        f"given: {configured_fields}\n"
                        f"found: {fieldnames}"
                    )
            else:
                self.field_type_map = {field: FieldType.str for field in fieldnames}
                LOG.info(f"No headers supplied for {self.name}, found {fieldnames}")

        else:
            self.fieldnames = self.field_type_map.keys()
        try:
            row = next(self.reader)
        except StopIteration:
            LOG.info(f"Finished processing {self.line_num} rows for {self.name}")
            raise StopIteration
        self.line_num = self.reader.line_num

        # skip blank lines
        if self.skip_blank_lines:
            while not row:
                row = next(self.reader)

        # Check row length discrepancies for each row
        # TODO currently varying line lengths will raise an exception
        # and hard fail, we should probably make these warnings and report
        # out which lines vary
        # Could also create a custom exception and allow the client code
        # to determine what to do here
        fields_len = len(self.fieldnames)
        row_len = len(row)
        if fields_len > row_len:
            LOG.warning(
                f"CSV file {self.name} has {fields_len - row_len} fewer columns at {self.reader.line_num}"
            )
        elif row_len > fields_len:
            LOG.warning(
                f"CSV file {self.name} has {row_len - fields_len} extra columns at {self.reader.line_num}"
            )

        # if we've made it here we can convert a row to a dict
        field_map = dict(zip(self.fieldnames, row))

        typed_field_map = {}

        for field, field_value in field_map.items():
            # This is really unreadable - malkovich malkovich
            # Take the value and coerce it using self.field_type_map (field: FieldType)
            # FIELD_TYPE is map of the field_type enum to the python
            # built-in type or custom extras defined in koza
            try:
                typed_field_map[field] = FIELDTYPE_CLASS[self.field_type_map[field]](field_value)
            except KeyError as key_error:
                LOG.warning(key_error)

        return typed_field_map



"""
source config data class
map config data class
"""
from enum import Enum
from pathlib import Path
from typing import Dict, List, Union

from pydantic import StrictFloat, StrictInt, StrictStr
from pydantic.dataclasses import dataclass

from koza.model.config.pydantic_config import PydanticConfig


class MapErrorEnum(str, Enum):
    """
    Enum for how to handle key errors in map files
    """

    warning = 'warning'
    error = 'error'


class FormatType(str, Enum):
    """
    Enum for supported file types
    """

    csv = 'csv'
    jsonl = 'jsonl'
    json = 'json'


class StandardFormat(str, Enum):
    gpi = 'gpi'
    bgi = 'bgi'
    oban = 'oban'


class CompressionType(str, Enum):
    """
    Enum for supported compression
    """

    gzip = 'gzip'
    none = 'none'


class FilterCode(str, Enum):
    """
    Enum for filter codes
    eg gt (greater than)
    """

    gt = 'gt'
    ge = 'ge'
    lt = 'lt'
    lte = 'le'
    eq = 'eq'
    ne = 'ne'
    inlist = 'in'


class FilterInclusion(str, Enum):
    """
    Enum for filter inclusion/exclusion
    """

    include = 'include'
    exclude = 'exclude'


class FieldType(str, Enum):
    """
    Enum for filter codes
    eg gt (greater than)
    """

    str = 'str'
    int = 'int'
    float = 'float'
    # Proportion = 'Proportion'


class OutputFormat(str, Enum):
    """
    Have this set up but for prototyping removing this
    as an option to only support the TSV output format
    """

    tsv = 'tsv'
    json = 'json'
    jsonl = 'jsonl'


class TransformMode(str, Enum):
    """
    Have this set up but for prototyping removing this
    as an option to only support the TSV output format
    """

    flat = 'flat'
    loop = 'loop'


@dataclass(frozen=True)
class ColumnFilter:
    column: str
    inclusion: FilterInclusion
    filter_code: FilterCode
    value: Union[StrictInt, StrictFloat, StrictStr, List[Union[StrictInt, StrictFloat, StrictStr]]]


@dataclass(frozen=True)
class DatasetDescription:
    id: str = None  # TODO constrain to a curie?
    name: str = None  # If empty use source name
    ingest_title: str = None  # Map to biolink name
    ingest_url: str = None  # Maps to biolink iri
    description: str = None
    source: str = None
    provided_by: str = None
    license: str = None
    rights: str = None


@dataclass(config=PydanticConfig)
class SourceConfig:
    source_files: List[str]
    name: str = None
    output_format: OutputFormat = None
    dataset_description: DatasetDescription = None


@dataclass(config=PydanticConfig)
class SourceFileConfig:
    """
    Base class for primary sources and mapping sources

    TODO document fields

    TODO override translation table path?

    delimiter:
    separator string similar to what works in str.split()
    https://docs.python.org/3/library/stdtypes.html#str.split

    required_properties: A list of required top level properties in a json object
    """

    name: str
    files: List[Union[str, Path]]
    format: FormatType = FormatType.csv
    path: Path = None
    standard_format: StandardFormat = None
    file_metadata: DatasetDescription = None
    columns: List[Union[str, Dict[str, FieldType]]] = None
    node_properties: List[str] = None
    edge_properties: List[str] = None
    required_properties: List[str] = None
    delimiter: str = None
    header_delimiter: str = None
    has_header: bool = True
    skip_lines: int = 0
    skip_blank_lines: bool = True
    compression: CompressionType = None
    filters: List[ColumnFilter] = None
    json_path: List[Union[StrictStr, StrictInt]] = None
    transform_code: str = None
    transform_mode: TransformMode = TransformMode.flat

    def __post_init_post_parse__(self):
        files_as_paths: List[Path] = []
        for file in self.files:
            if isinstance(file, str):
                files_as_paths.append(Path(file))
            else:
                files_as_paths.append(file)
        object.__setattr__(self, 'files', files_as_paths)

        # todo: where should this really be stored? defaults for a format should probably be defined in yaml
        if self.standard_format == StandardFormat.gpi:
            self.format = FormatType.csv
            self.delimiter = "\t"
            self.columns = [
                "DB",
                "DB_Object_ID",
                "DB_Object_Symbol",
                "DB_Object_Name",
                "DB_Object_Synonym(s)",
                "DB_Object_Type",
                "Taxon",
                "Parent_Object_ID",
                "DB_Xref(s)",
                "Properties",
            ]
        elif self.standard_format == StandardFormat.oban:
            self.format = FormatType.csv
            self.delimiter = ","
            self.columns = [
                "SUBJECT",
                "SUBJECT_LABEL",
                "SUBJECT_TAXON",
                "SUBJECT_TAXON_LABEL",
                "OBJECT",
                "OBJECT_LABEL",
                "RELATION",
                "RELATION_LABEL",
                "EVIDENCE",
                "EVIDENCE_LABEL",
                "SOURCE",
                "IS_DEFINED_BY",
                "QUALIFIER",
            ]

        if self.delimiter in ['tab', '\\t']:
            object.__setattr__(self, 'delimiter', '\t')

        if self.filters is None:
            self.filters = []

        filtered_columns = [column_filter.column for column_filter in self.filters]

        all_columns = []
        if self.columns:
            all_columns = [
                next(iter(column)) if isinstance(column, Dict) else column
                for column in self.columns
            ]

        for column in filtered_columns:
            if column not in all_columns:
                raise (ValueError(f"Filter column {column} not in column list"))

        for column_filter in self.filters:
            if column_filter.filter_code in ['lt', 'gt', 'lte', 'gte']:
                # TODO determine if this should raise an exception
                # or instead try to type coerce the string to a float
                # type coercion is probably the best thing to do here
                if not isinstance(column_filter.value, (int, float)):
                    raise ValueError(
                        f"Filter value must be int or float for operator {column_filter.filter_code}"
                    )
            elif column_filter.filter_code == 'eq':
                if not isinstance(column_filter.value, (str, int, float)):
                    raise ValueError(
                        f"Filter value must be string, int or float for operator {column_filter.filter_code}"
                    )
            elif column_filter.filter_code == 'in':
                if not isinstance(column_filter.value, List):
                    raise ValueError(
                        f"Filter value must be List for operator {column_filter.filter_code}"
                    )

        if self.format == FormatType.csv and self.required_properties:
            raise ValueError(
                "csv specified but required properties have been configured\n"
                "either set format to jsonl or change properties to columns in the config"
            )

        if self.columns and self.format != FormatType.csv:
            raise ValueError(
                "columns have been configured but format is not csv\n"
                "either set format to csv or change columns to properties in the config"
            )

        if self.json_path and self.format != FormatType.json:
            raise ValueError(
                "iterate_over has been configured but format is not json\n"
                "either set format to json or remove iterate_over in the configuration"
            )

        if self.columns:
            _field_type_map = {}
            for field in self.columns:
                if isinstance(field, str):
                    _field_type_map[field] = FieldType.str
                else:
                    if len(field) != 1:
                        # TODO expand this exception msg
                        raise ValueError("Field type map contains more than one key")
                    for key, val in field.items():
                        _field_type_map[key] = val
            object.__setattr__(self, '_field_type_map', _field_type_map)

    @property
    def field_type_map(self):
        return self._field_type_map


@dataclass(config=PydanticConfig)
class PrimaryFileConfig(SourceFileConfig):
    depends_on: List[str] = None  # field(default_factory=list)
    on_map_failure: MapErrorEnum = MapErrorEnum.warning


@dataclass(config=PydanticConfig)
class MapFileConfig(SourceFileConfig):
    source: str = None
    key: str = None
    values: List[str] = None
    curie_prefix: str = None
    add_curie_prefix_to_columns: List[str] = None



import logging
import re

from koza.manager.data_provider import inject_row, inject_translation_table, inject_map

LOG = logging.getLogger(__name__)

source_name = 'mimtitles'

row = inject_row(source_name)
translation_table = inject_translation_table(source_name)
map = inject_map(source_name)

entry = dict()

###
# From OMIM
#  An asterisk (*) before an entry number indicates a gene.
#
# A number symbol (#) before an entry number indicates that it is a descriptive entry,
# usually of a phenotype, and does not represent a unique locus. The reason for the use
# of the number symbol is given in the first paragraph of he entry. Discussion of any gene(s)
# related to the phenotype resides in another entry(ies) as described in the first paragraph.
#
# A plus sign (+) before an entry number indicates that the entry contains the description of
# a gene of known sequence and a phenotype.
#
# A percent sign (%) before an entry number indicates that the entry describes a confirmed
# mendelian phenotype or phenotypic locus for which the underlying molecular basis is not known.
#
# No symbol before an entry number generally indicates a description of a phenotype for which
# the mendelian basis, although suspected, has not been clearly established or that the separateness
# of this phenotype from that in another entry is unclear.
#
# A caret (^) before an entry number means the entry no longer exists because it was removed
# from the database or moved to another entry as indicated.
###

if row['Prefix'] == 'Asterisk':
    entry['type'] = translation_table.global_table['gene']

elif row['Prefix'] == 'NULL':
    entry['type'] = translation_table.global_table['Suspected']  # NCIT:C71458

elif row['Prefix'] == 'Number Sign':
    entry['type'] = translation_table.global_table['phenotype']

elif row['Prefix'] == 'Percent':
    entry['type'] = translation_table.global_table['heritable_phenotypic_marker']

elif row['Prefix'] == 'Plus':
    entry['type'] = translation_table.global_table['gene']

elif row['Prefix'] == 'Caret':  # moved|removed|split -> moved twice
    # populating a dict from an omim to a set of omims
    entry['type'] = translation_table.global_table['obsolete']
    entry['replaced by'] = []

    if row['Preferred Title; symbol'][:9] == 'MOVED TO ':
        token = row['Preferred Title; symbol'].split(' ')
        title_symbol = token[2]

        if not re.match(r'^[0-9]{6}$', title_symbol):
            LOG.error(f"Report malformed omim replacement {title_symbol}")
            # clean up ones I know about

            if title_symbol[0] == '{' and title_symbol[7] == '}':
                title_symbol = title_symbol[1:7]
                LOG.info(f"Repaired malformed omim replacement {title_symbol}")

            if len(title_symbol) == 7 and title_symbol[6] == ',':
                title_symbol = title_symbol[:6]
                LOG.info(f"Repaired malformed omim replacement {title_symbol}")

        if len(token) > 3:
            entry['replaced by'] = [title_symbol, token[4]]

        else:
            entry['replaced by'] = [title_symbol]

else:
    LOG.error(f"Unknown OMIM type line {row['omim_id']}")

map[row['MIM Number']] = entry



import logging
import uuid

from koza.manager.data_provider import inject_row, inject_map, inject_translation_table
from koza.manager.data_collector import write
from biolink_model_pydantic.model import Publication, Gene, NamedThingToInformationContentEntityAssociation, Predicate

LOG = logging.getLogger(__name__)

source_name = 'gene-literature'

row = inject_row(source_name)
translation_table = inject_translation_table()
genepage2gene = inject_map('genepage-2-gene')

entities = []

gene_pages = row['gene_pages']

publication = Publication(
    id='PMID:' + row['pmid'],
    type=translation_table.resolve_term("publication")
)

entities.append(publication)

for gene_page in gene_pages.split(','):
    gene_page_id = gene_page.split(' ')[0]
    try:
        gene_ids = map(lambda id: f"Xenbase:{id}", list(genepage2gene[gene_page_id].values()))
    except KeyError:
        LOG.debug(f"Could not locate genepage_id: {gene_page_id} in row {row}")
        continue
    for gene_id in gene_ids:
        gene = Gene(id=gene_id)

        entities.append(gene)

        association = NamedThingToInformationContentEntityAssociation(
            id="uuid:" + str(uuid.uuid1()),
            subject=gene.id,
            predicate=Predicate.mentions,
            object=publication.id,
            relation="IAO:0000142"  # Mentions
        )

        entities.append(association)

write(source_name, *entities)



import pytest


@pytest.fixture
def gene_information_entities(mock_koza):
    row = iter(
        [
            {
                'DB': 'Xenbase',
                'DB_Object_ID': 'XB-GENE-478054',
                'DB_Object_Symbol': 'trnt1',
                'DB_Object_Name': 'tRNA nucleotidyl transferase, CCA-adding, 1',
                'DB_Object_Synonym(s)': 'cca1|mtcca|tRNA nucleotidyl transferase|xcca',
                'DB_Object_Type': 'gene',
                'Taxon': 'taxon:8364',
                'Parent_Object_ID': '',
                'DB_Xref(s)': 'NCBI_Gene:394602|UniProtKB:F6V8Z1|UniProtKB:F7DM70|UniProtKB:Q6P873',
                'Properties': '',
            }
        ]
    )

    return mock_koza('gene-information', row, './examples/xenbase/gene-information.py')


def test_gene_information_gene(gene_information_entities):
    assert len(gene_information_entities) == 1
    gene = gene_information_entities[0]
    assert gene


def test_gene_information_synonym(gene_information_entities):
    gene = gene_information_entities[0]
    assert gene.synonym
    assert 'xcca' in gene.synonym


def test_gene_information_id(gene_information_entities):
    gene = gene_information_entities[0]
    assert gene.id == 'Xenbase:XB-GENE-478054'



from biolink_model_pydantic.model import Gene

from koza.manager.data_provider import inject_curie_cleaner


def gpi2gene(row: dict) -> Gene:
    """
    Convert from Gene Product Information format to biolink:Gene instance

    http://geneontology.org/docs/gene-product-information-gpi-format/

    :param row: Dictionary representing a single GPI file row
    :return: biolink:Gene model representing the GPI row
    """

    curie_cleaner = inject_curie_cleaner()

    xrefs = []
    if row["DB_Xref(s)"]:
        xrefs = [curie_cleaner.clean(xref) for xref in row["DB_Xref(s)"].split("|")]

    gene = Gene(
        id=row['DB_Object_ID'],
        symbol=row['DB_Object_Symbol'],
        name=row['DB_Object_Name'],
        synonym=row['DB_Object_Synonym(s)'].split("|") if row['DB_Object_Synonym(s)'] else [],
        in_taxon=curie_cleaner.clean(row['Taxon']),
        xref=xrefs,
        source=row['DB'],
    )

    return gene



from typing import Dict

from koza.koza_runner import get_koza_app
from koza.model.curie_cleaner import CurieCleaner
from koza.model.translation_table import TranslationTable


def inject_row(ingest_name: str) -> Dict:
    koza = get_koza_app()
    if ingest_name in koza.file_registry:
        return next(koza.file_registry[ingest_name])
    elif ingest_name in koza.map_registry:
        return next(koza.map_registry[ingest_name])
    else:
        raise KeyError(f"{ingest_name} not found in file or map registry")


def inject_map(map_name: str) -> Dict:
    """
    Get map associated with the specified source
    :param map_name:
    :param source_name:
    :return:
    """
    koza = get_koza_app()
    return koza.map_cache[map_name]


def inject_translation_table() -> TranslationTable:
    """"""
    koza = get_koza_app()
    return koza.source.translation_table


def inject_curie_cleaner() -> CurieCleaner:
    koza = get_koza_app()
    return koza.curie_cleaner



"""
Package for IO modules - reader, writer, serializer

Note that naming this package 'io' breaks the pycharm debugger
https://youtrack.jetbrains.com/issue/PY-23615

A workaround is to remove the working directory in the run/debug configuration
"""



"""
Stub for testing koza.io.utils

Can use the io lib to mock files, or the mock lib
https://docs.python.org/3/library/io.html
https://docs.python.org/3/library/unittest.mock.html

https://github.com/monarch-initiative/dipper/blob/6f1fe5bb/dipper/utils/TestUtils.py#L41
https://github.com/monarch-initiative/dipper/blob/682560f/tests/test_udp.py#L85
"""

import pytest

from koza.io.utils import open_resource


def test_404():
    resource = "http://httpstat.us/404"
    with pytest.raises(ValueError):
        with open_resource(resource) as _:
            pass



import logging

from koza.converter.biolink_converter import gpi2gene
from koza.manager.data_provider import inject_row
from koza.manager.data_collector import write

LOG = logging.getLogger(__name__)

source_name = 'gene-information'
row = inject_row(source_name)

row['DB_Object_ID'] = "Xenbase:" + row['DB_Object_ID']

gene = gpi2gene(row)

write(source_name, gene)



"""
Custom exceptions
"""


class NextRowException(Exception):
    """Triggers an iterator to continue to the next row"""


class MapItemException(KeyError):
    """
    Special case of KeyError for source maps based on configuration,
    a source may opt to warn or exit with an error
    """



import uuid

from biolink_model_pydantic.model import Gene, PhenotypicFeature, GeneToPhenotypicFeatureAssociation, Predicate
from koza.manager.data_provider import inject_row
from koza.manager.data_collector import write

source_name = 'gene-to-phenotype'

row = inject_row(source_name)

gene = Gene(id='Xenbase:' + row['SUBJECT'])

phenotype = PhenotypicFeature(id=row['OBJECT'])

association = GeneToPhenotypicFeatureAssociation(
    id="uuid:" + str(uuid.uuid1()),
    subject=gene.id,
    predicate=Predicate.has_phenotype,
    object=phenotype.id,
    relation=row['RELATION'].replace('_', ':')
)

if row['SOURCE']:
    association.publications = [row['SOURCE']]

write(source_name, gene, association, phenotype)



import re


class CurieCleaner:
    def __init__(self):
        # TODO: these belong in yaml, might already exist somewhere
        self.mappings = {"taxon": "NCBITaxon", "NCBI_Gene": "NCBIGene"}

    def clean(self, curie: str) -> str:
        for curie_synonym in self.mappings:
            curie = re.sub("^%s:" % curie_synonym, self.mappings[curie_synonym] + ":", curie)
        return curie



import gzip
from pathlib import Path

import pytest

from koza.io.reader.json_reader import JSONReader

test_ddpheno = Path(__file__).parents[1] / 'resources' / 'source-files' / 'ddpheno.json.gz'

json_path = ['graphs', 0, 'nodes']


def test_normal_case():
    with gzip.open(test_ddpheno, 'rt') as ddpheno:
        json_reader = JSONReader(ddpheno, json_path=json_path)
        row = next(json_reader)
        assert row['id'] == 'http://purl.obolibrary.org/obo/DDPHENO_0001198'


def test_required_properties():
    with gzip.open(test_ddpheno, 'rt') as ddpheno:
        json_reader = JSONReader(ddpheno, ['id'], json_path=json_path)
        for row in json_reader:
            assert 'id' in row


def test_missing_req_property_raises_exception():
    with gzip.open(test_ddpheno, 'rt') as ddpheno:
        json_reader = JSONReader(ddpheno, ['fake_prop'], json_path=json_path)
        with pytest.raises(ValueError):
            next(json_reader)





global_translation = {}
file = {}
# Nodes

#hp_term = Node
#publications: List[Publication] = []
# #disease = Disease


for pub in file['Publications'].split(';'):

    publication.id = pub

    if pub.startswith('PMID:'):
        publication.type = global_translation['journal article']

    elif pub.startswith('ISBN'):
        publication.type = global_translation['publication']

    elif pub.startswith('OMIM:'):
        publication.type = global_translation['web page']

    elif pub.startswith('DECIPHER:'):
        publication.type = global_translation['web page']

    elif pub.startswith('ORPHA:'):
        publication.type = global_translation['web page']

    elif pub.startswith('http'):
        publication.type = global_translation['web page']

    else:
        pass
        # not mapped warning

# Edges

if file['Aspect'] == 'P':
    pass

elif file['Aspect'] == 'M':
    pass

elif file['Aspect'] == 'I':
    pass

elif file['Aspect'] == 'C':
    pass


import uuid

from biolink_model_pydantic.model import Gene, PairwiseGeneToGeneInteraction, Predicate
from koza.manager.data_provider import inject_row, inject_translation_table, inject_map
from koza.manager.data_collector import write

source_name = 'protein-links-detailed'

row = inject_row(source_name)
translation_table = inject_translation_table()
entrez_2_string = inject_map('custom_entrez_2_string')

gene_a = Gene(id='NCBIGene:' + entrez_2_string[row['protein1']]['entrez'])
gene_b = Gene(id='NCBIGene:' + entrez_2_string[row['protein2']]['entrez'])

pairwise_gene_to_gene_interaction = PairwiseGeneToGeneInteraction(
    id="uuid:" + str(uuid.uuid1()),
    subject=gene_a.id,
    object=gene_b.id,
    predicate=Predicate.interacts_with,
    relation = translation_table.global_table['interacts with'],
)

write(source_name, gene_a, gene_b, pairwise_gene_to_gene_interaction)



import re
import uuid

from biolink_model_pydantic.model import Protein, PairwiseGeneToGeneInteraction, Predicate
from koza.manager.data_provider import inject_row, inject_translation_table
from koza.manager.data_collector import write

source_name = 'protein-links-detailed'

row = inject_row(source_name)
translation_table = inject_translation_table()

protein_a = Protein(id='ENSEMBL:' + re.sub(r'\d+\.', '', row['protein1']))
protein_b = Protein(id='ENSEMBL:' + re.sub(r'\d+\.', '', row['protein2']))

pairwise_gene_to_gene_interaction = PairwiseGeneToGeneInteraction(
    id="uuid:" + str(uuid.uuid1()),
    subject=protein_a.id,
    object=protein_b.id,
    predicate=Predicate.interacts_with,
    relation = translation_table.global_table['interacts with'],
)

write(source_name, protein_a, protein_b, pairwise_gene_to_gene_interaction)



"""
Functions that collect biolink model objects and write them
"""

from ..koza_runner import get_koza_app


def write(source_name, *args):
    koza = get_koza_app()
    koza.write(source_name, list(args))



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['koza',
 'koza.converter',
 'koza.io',
 'koza.io.reader',
 'koza.io.writer',
 'koza.manager',
 'koza.model',
 'koza.model.config']

package_data = \
{'': ['*']}

install_requires = \
['biolink_model_pydantic >=0.0.5,<0.1.0',
 'kgx >= 1.0.0',
 'pydantic >=1.0.0,<2.0.0',
 'pyyaml >=5.3.1,<6.0.0',
 'requests >=2.24.0,<3.0.0',
 'typer >=0.3']

extras_require = \
{'dev': ['autoflake >=1.3.1,<2.0.0',
         'flake8 >=3.8.3,<4.0.0',
         'black ==20.8b1',
         'isort >=5.0.6,<6.0.0'],
 'doc': ['mkdocs >=1.1.2,<2.0.0'],
 'test': ['pytest >=6.0.0']}

entry_points = \
{'console_scripts': ['koza = koza:main.app']}

setup(name='koza',
      version='0.0.1a5',
      description='Koza, an ETL framework for the Biolink model',
      author='The Monarch Initiative',
      author_email='info@monarchinitiative.org',
      url='https://github.com/monarch-initiative/koza',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



from koza.model import Gene, PairwiseGeneToGeneInteraction, predicate
from koza.manager.data_provider import inject_file, inject_translation_table
from koza.manager.data_collector import serialize
from koza.manager.helper import next_row

_ingest_name = 'protein-links-detailed'

file, entrez_2_string, = inject_file(_ingest_name)
translation_table = inject_translation_table(_ingest_name)

gene_a = Gene()  # Maps to file['protein1']
gene_b = Gene()  # Maps to file['protein2']

pairwise_gene_to_gene_interaction = PairwiseGeneToGeneInteraction(provided_by='stringdb')

model = (gene_a, gene_b, pairwise_gene_to_gene_interaction)


serialize(_ingest_name, *model)



from pathlib import Path

import pytest

from koza.io.reader.csv_reader import CSVReader
from koza.model.config.source_config import FieldType

test_file = Path(__file__).parent.parent / 'resources' / 'source-files' / 'string.tsv'

field_type_map = {
    'protein1': FieldType.str,
    'protein2': FieldType.str,
    'neighborhood': FieldType.str,
    'fusion': FieldType.str,
    'cooccurence': FieldType.str,
    'coexpression': FieldType.str,
    'experimental': FieldType.str,
    'database': FieldType.str,
    'textmining': FieldType.float,
    'combined_score': FieldType.int,
}


def test_no_exceptions_in_normal_case():
    with open(test_file, 'r') as string_file:
        reader = CSVReader(string_file, field_type_map, delimiter=' ')
        # TODO actually test something
        for row in reader:
            pass


def test_type_conversion():
    with open(test_file, 'r') as string_file:
        reader = CSVReader(string_file, field_type_map, delimiter=' ')
        row = next(reader)
        assert isinstance(row['protein1'], str)
        assert isinstance(row['textmining'], float)
        assert isinstance(row['combined_score'], int)


def test_field_doesnt_exist_in_file_raises_exception():
    with open(test_file, 'r') as string_file:
        field_map = field_type_map.copy()
        field_map['some_field_that_doesnt_exist'] = FieldType.str
        reader = CSVReader(string_file, field_map, delimiter=' ')
        with pytest.raises(ValueError):
            next(reader)


def test_field_in_file_but_not_in_config_logs_warning(caplog):
    """
    https://docs.pytest.org/en/latest/logging.html#caplog-fixture
    :return:
    """
    with open(test_file, 'r') as string_file:
        field_map = field_type_map.copy()
        del field_map['combined_score']
        reader = CSVReader(string_file, field_map, delimiter=' ')
        next(reader)
        assert caplog.records[0].levelname == 'WARNING'
        assert caplog.records[0].msg.startswith('Additional column(s) in source file')


def test_middle_field_in_file_but_not_in_config_logs_warning(caplog):
    with open(test_file, 'r') as string_file:
        field_map = field_type_map.copy()
        del field_map['cooccurence']
        reader = CSVReader(string_file, field_map, delimiter=' ')
        next(reader)
        assert caplog.records[1].levelname == 'WARNING'
        assert caplog.records[1].msg.startswith(
            'Additional columns located within configured fields'
        )


def test_no_field_map(caplog):
    with open(test_file, 'r') as string_file:
        reader = CSVReader(string_file, delimiter=' ')
        next(reader)



#!/usr/bin/env python3

"""
A pydantic generator based on the LinkML python generator

Some key differences:

- pydantic dataclasses instead of vanilla dataclasses,
  see https://pydantic-docs.helpmanual.io/usage/dataclasses/

- PredicateType is replaced by a Predicate enum

- UriOrCurie is replaced with a Curie type as a goal to represent
  all identifiers as curies, IriType is still included in a few attributes

- Identifier types are removed, eg Union[str, EntityId] is replaced with
  pydantic constr (Curie) constrained by a valid curie regex
  https://www.w3.org/TR/curie/
  https://pydantic-docs.helpmanual.io/usage/types/#arguments-to-constr

- Category attribute is inferred via class variables and the type hierarchy
    - Note that for id and type, and sometimes other attributes these
      are overridden anyway in the child class.  I think this happens when
      a slot if given a new description in the child class, linkml gives
      the a slot a unique name (child_class_id, child_class_type) to
      attach the updated description, need to check this with Harold

- Type conversions, converts scalars to lists for Union[someScalar, List[someScalar]]


Downstream code will need to handle nested types to be compliant with
Neo4J's data model.  Nested types will need to be converted to some primitive type
(string, number, or lists of a primitive type)


Why pydantic over standard dataclasses?

  - Validation on both initializing and setting of variables

  - Built in type coercion (this is perhaps a con as Union types are handled in odd ways for now)
    see https://github.com/samuelcolvin/pydantic/issues/1423
    https://github.com/samuelcolvin/pydantic/pull/2092

  - Built in parsing of json or yaml into nested models (ie when attributes are reference types)

"""

from pathlib import Path
from typing import List, Optional, TextIO, Tuple, Union

import typer
from linkml.generators import PYTHON_GEN_VERSION
from linkml.generators.pythongen import PythonGenerator
from linkml.utils.formatutils import be, camelcase, split_line, wrapped_annotation
from linkml_model.meta import ClassDefinition, ClassDefinitionName, SchemaDefinition, SlotDefinition


class PydanticGen(PythonGenerator):
    """
    A pydantic dataclass generator

    """

    generatorname = Path(__file__).name
    generatorversion = PYTHON_GEN_VERSION

    def __init__(self, schema: Union[str, TextIO, SchemaDefinition], **kwargs):
        super().__init__(
            schema=schema,
            format='py',
            genmeta=False,
            gen_classvars=False,
            gen_slots=False,
            **kwargs,
        )

    # Overriden methods

    def _sort_classes(self, clist: List[ClassDefinition]) -> List[ClassDefinition]:
        """
        sort classes such that if C is a child of P then C appears after P in the list

        Overridden method include mixin classes
        """
        clist = list(clist)
        slist = []  # sorted
        while len(clist) > 0:
            can_add = False
            for i in range(len(clist)):
                candidate = clist[i]
                can_add = False
                if candidate.is_a:
                    candidates = [candidate.is_a] + candidate.mixins
                else:
                    candidates = candidate.mixins
                if not candidates:
                    can_add = True
                else:
                    if set(candidates) <= set([p.name for p in slist]):
                        can_add = True
                if can_add:
                    slist = slist + [candidate]
                    del clist[i]
                    break
            if not can_add:
                raise ValueError(
                    f'could not find suitable element in {clist} that does not ref {slist}'
                )
        return slist

    def gen_schema(self) -> str:
        split_descripton = '\n#              '.join(
            split_line(be(self.schema.description), split_len=100)
        )
        head = (
            f'''# Auto generated from {self.schema.source_file} by {self.generatorname} version: {self.generatorversion}
# Generation date: {self.schema.generation_date}
# Schema: {self.schema.name}
#'''
            if self.schema.generation_date
            else ''
        )

        return f'''{head}
# id: {self.schema.id}
# description: {split_descripton}
# license: {be(self.schema.license)}

import datetime
import inspect
import logging
from enum import Enum
from dataclasses import field
from typing import Any, ClassVar, List, Optional, Union
import re

from pydantic import constr, validator
from pydantic.dataclasses import dataclass

LOG = logging.getLogger(__name__)

metamodel_version = "{self.schema.metamodel_version}"

curie_regexp = r'^[a-zA-Z_]?[a-zA-Z_0-9-]*:([A-Za-z0-9_][A-Za-z0-9_.-]*[A-Za-z0-9_]*)?$'
curie_pattern = re.compile(curie_regexp)

# Type Aliases
Unit = Union[int, float]
LabelType = str
IriType = constr(regex=r'^http')
Curie = constr(regex=curie_regexp)
NarrativeText = str
XSDDate = datetime.date
TimeType = datetime.time
SymbolType = str
FrequencyValue = str
PercentageFrequencyValue = float
BiologicalSequence = str
Quotient = float

# Namespaces
{self.gen_namespaces()}

# Pydantic config and validators
{self._get_pydantic_config_and_validators()}

# Predicates
{self.gen_predicate_enum()}

# Classes
{self.gen_classdefs()}

'''

    def gen_classdef(self, cls: ClassDefinition) -> str:
        """ Generate python definition for class cls """

        parent_class_and_mixins = ""

        if cls.is_a:
            parents = [cls.is_a]
            if cls.mixins:
                parents = parents + cls.mixins
            parent_class_and_mixins = ', '.join(
                [self.formatted_element_name(parent, True) for parent in parents]
            )
            parent_class_and_mixins = f'({parent_class_and_mixins})'
        elif cls.mixins:
            # Seems fine but more curious if this ever happens
            self.logger.warning(f"class {cls.name} has mixins {cls.mixins} but no parent")
            mixins = ', '.join([self.formatted_element_name(mixin, True) for mixin in cls.mixins])
            parent_class_and_mixins = f'({mixins})'

        slotdefs = self.gen_class_variables(cls)

        entity_post_init = (
            f'\n\t{self._get_entity_post_init()}'
            if self.class_or_type_name(cls.name) == 'Entity'
            else ''
        )

        wrapped_description = (
            f'\n\t"""\n\t{wrapped_annotation(be(cls.description))}\n\t"""'
            if be(cls.description)
            else ''
        )

        pydantic_classvars = self.gen_pydantic_classvars(cls)

        pydantic_validators = self.gen_pydantic_validators(cls)

        return (
            ('\n@dataclass(config=PydanticConfig)')
            + f'\nclass {self.class_or_type_name(cls.name)}'
            + parent_class_and_mixins
            + f':{wrapped_description}\n'
            + f'{self.gen_inherited_slots(cls)}'
            + ('\n\t# Class Variables' if pydantic_classvars else '')
            + f'{pydantic_classvars}'
            + (f'\t{slotdefs}\n' if slotdefs else '')
            + f'{pydantic_validators}'
            + f'{entity_post_init}'
            + (
                f'\n\tpass'
                if (
                    not self.gen_inherited_slots(cls)
                    and not pydantic_classvars
                    and not slotdefs
                    and not entity_post_init
                    and not pydantic_validators
                )
                else ''
            )
            + '\n'
        )

    def gen_class_variables(self, cls: ClassDefinition) -> str:
        """
        Generate the variable declarations for a dataclass.

        Overriden to only generate variables for domain slots

        :param cls: class containing variables to be rendered in inheritence hierarchy
        :return: variable declarations for target class and its ancestors
        """
        initializers = []

        domain_slots = self.domain_slots(cls)

        # Required or key slots with default values
        slot_variables = self._slot_iter(cls, lambda slot: slot.required and slot in domain_slots)
        initializers += [self.gen_class_variable(cls, slot, False) for slot in slot_variables]

        # Followed by everything else
        slot_variables = self._slot_iter(
            cls, lambda slot: not slot.required and slot in domain_slots
        )
        initializers += [self.gen_class_variable(cls, slot, False) for slot in slot_variables]

        return '\n\t'.join(initializers)

    def range_cardinality(
        self, slot: SlotDefinition, cls: Optional[ClassDefinition], positional_allowed: bool
    ) -> Tuple[str, Optional[str]]:
        """
        Overriding to switch empty_list() and empty_dict() to

        field(default_factory={list | dict})
        """
        positional_allowed = False  # Force everything to be tag values
        slotname = self.slot_name(slot.name)

        range_type, parent_type, _ = self.class_reference_type(slot, cls)
        pkey = self.class_identifier(slot.range)
        # Special case, inlined, identified range
        if pkey and slot.inlined and slot.multivalued:
            base_key = self.gen_class_reference(
                self.class_identifier_path(slot.range, False), slotname
            )
            num_elements = len(self.schema.classes[slot.range].slots)
            dflt = None if slot.required and positional_allowed else 'field(default_factory=dict)'
            if num_elements == 1:
                if slot.required:
                    return f'Union[List[{base_key}], Dict[{base_key}, {range_type}]]', dflt
                else:
                    return (
                        f'Optional[Union[List[{base_key}], Dict[{base_key}, {range_type}]]]',
                        dflt,
                    )
            else:
                if slot.required:
                    return f'Union[Dict[{base_key}, {range_type}], List[{range_type}]]', dflt
                else:
                    return (
                        f'Optional[Union[Dict[{base_key}, {range_type}], List[{range_type}]]]',
                        dflt,
                    )

        # All other cases
        if slot.multivalued:
            if slot.required:
                return f'Union[{range_type}, List[{range_type}]]', (
                    None if positional_allowed else 'None'
                )
            else:
                return (
                    f'Optional[Union[{range_type}, List[{range_type}]]]',
                    'field(default_factory=list)',
                )
        elif slot.required:
            return range_type, (None if positional_allowed else 'None')
        else:
            return f'Optional[{range_type}]', 'None'

    def class_reference_type(
        self, slot: SlotDefinition, cls: Optional[ClassDefinition]
    ) -> Tuple[str, str, str]:
        """
        Return the type of a slot referencing a class

        :param slot: slot to be typed
        :param cls: owning class.  Used for generating key references
        :return: Python class reference type, most proximal type, most proximal type name
        """

        slotname = self.slot_name(slot.name)

        rangelist = (
            self.class_identifier_path(cls, False)
            if slot.key or slot.identifier
            else self.slot_range_path(slot)
        )
        prox_type = self.slot_range_path(slot)[-1].rsplit('.')[-1]
        prox_type_name = rangelist[-1]

        # Quote forward references - note that enums always gen at the end
        if slot.range in self.schema.enums or (
            cls
            and slot.inlined
            and slot.range in self.schema.classes
            and self.forward_reference(slot.range, cls.name)
        ):
            rangelist[-1] = f'"{rangelist[-1]}"'

        # return str(self.gen_class_reference(rangelist)), prox_type, prox_type_name
        return str(self.gen_class_reference(rangelist, slotname)), prox_type, prox_type_name

    @staticmethod
    def gen_class_reference(rangelist: List[str], slot_name: str = None) -> str:
        """
        Return a basic or a union type depending on the number of elements in range list

        Instead of the pythongen version which uses the base type and a special
        Id type, eg
        Entity -> str, EntityID

        We have a union of str, Curie, and the class, eg
        Entity -> str, Curie, Entity

        TODO clean this up

        :param rangelist: List of types from distal to proximal
        :return:
        """
        base = rangelist[0].split('.')[-1]

        if slot_name in ['category', 'provided_by']:
            class_ref = "Union[str, Curie]"
        elif slot_name == 'predicate':
            class_ref = "Predicate"
        elif (
            slot_name
            in [
                'id',
                'xref',
                'has_qualitative_value',
                'subclass_of',
                'has_input',
                'has_output',
                'has_constituent',
                'enabled_by',
            ]
            or ('URIorCURIE' in rangelist and rangelist[-1] != 'IriType')
        ):
            # id is here to override {ClassName}Id
            # The rest are due to import order errors in self._sort_classes
            # From changing {ClassName}Id to {ClassName}

            # Check if base is str? for now just assume it is
            class_ref = "Curie"

        elif 'Entity' in rangelist:
            class_ref = f"Union[Curie, {rangelist[-1]}]" if len(rangelist) > 1 else base
        elif 'Bool' == rangelist[-1]:
            class_ref = 'bool'
        else:
            if len(rangelist) > 1 and rangelist[-1] == 'IriType':
                class_ref = 'IriType'
            else:
                class_ref = f"Union[{base}, {rangelist[-1]}]" if len(rangelist) > 1 else base

        return class_ref

    def class_identifier_path(
        self, cls_or_clsname: Union[str, ClassDefinition], force_non_key: bool
    ) -> List[str]:
        """
        Return the path closure to a class identifier if the class has a key and force_non_key is false otherwise
        return a dictionary closure.

        :param cls_or_clsname: class definition
        :param force_non_key: True means inlined even if the class has a key
        :return: path
        """
        cls = (
            cls_or_clsname
            if isinstance(cls_or_clsname, ClassDefinition)
            else self.schema.classes[ClassDefinitionName(cls_or_clsname)]
        )

        # Determine whether the class has a key
        identifier_slot = None
        if not force_non_key:
            identifier_slot = self.class_identifier(cls)

        # No key or inlined, its closure is a dictionary
        if identifier_slot is None:
            # return ['dict', self.class_or_type_name(cls.name)]
            # Not certain why this is dict and if it's a model smell
            # We want everything to be str, Curie, List, or another Dataclass in the model
            return ['str', self.class_or_type_name(cls.name)]

        # Override class name + self.aliased_slot_name
        # For example, instead of EntityId, which means nothing for the pydantic gen
        # use the dataclass itself Entity
        # pathname = camelcase(cls.name + ' ' + self.aliased_slot_name(identifier_slot))
        pathname = camelcase(cls.name)
        if cls.is_a:
            parent_identifier_slot = self.class_identifier(cls.is_a)
            if parent_identifier_slot:
                return self.class_identifier_path(cls.is_a, False) + [pathname]
        return self.slot_range_path(identifier_slot) + [pathname]

    def domain_slots(self, cls: ClassDefinition) -> List[SlotDefinition]:
        """
        Return all slots in the class definition that are owned by the class

        Overridden to remove domain_of mixin(s) slots so we can use the mixin
        slot inheritance instead
        """
        return [
            slot
            for slot in [self.schema.slots[sn] for sn in cls.slots]
            if cls.name in slot.domain_of
        ]

    def gen_namespaces(self) -> str:

        namespaces = ',\n'.join(
            [
                f'\t"{pfx.replace(".", "_")}"'
                for pfx in sorted(self.emit_prefixes)
                if pfx in self.namespaces
            ]
        )

        return f'''
valid_prefix = [
{namespaces}
]
'''

    # New Methods

    def gen_pydantic_classvars(self, cls: ClassDefinition) -> str:
        """
        Generate classvars specific to the pydantic dataclasses
        """

        vars = []

        if not (cls.mixin or cls.abstract):
            vars.append(f'_category: ClassVar[str] = "{camelcase(cls.name)}"')

        id_prefixes_fmt = ',\n'.join([f'\t\t"{prefix}"' for prefix in cls.id_prefixes])

        if id_prefixes_fmt:
            vars.append(f'_id_prefixes: ClassVar[List[str]] = [\n{id_prefixes_fmt}\n\t]')

        if vars:
            ret_val = "\n\t" + "\n\t".join(vars) + "\n\n"
        else:
            ret_val = ""

        return ret_val

    def gen_predicate_enum(self) -> str:
        """
        Creates a named tuple of all biolink predicates
        which are slots that descend from 'related to'
        :return:
        """
        predicates = []
        for slot in self.schema.slots.values():
            if 'related to' in self.ancestors(slot):
                predicates.append(slot.name)

        predicates = [pred.replace(' ', '_') for pred in sorted(predicates)]
        formatted_predicates = '\n'.join([f'    {pred} = "biolink:{pred}"' for pred in predicates])

        return f'''
class Predicate(str, Enum):
    """
    Enum for biolink predicates
    """

{formatted_predicates}
'''

    def gen_pydantic_validators(self, cls) -> str:
        """

        :param cls:
        :return:
        """
        validators = []

        for slot in self.domain_slots(cls):
            slotname = self.slot_name(slot.name)
            rangelist = (
                self.class_identifier_path(cls, False)
                if slot.key or slot.identifier
                else self.slot_range_path(slot)
            )
            if (
                slotname
                in [
                    'id',
                    'xref',
                    'has_qualitative_value',
                    'subclass_of',
                    'has_input',
                    'has_output',
                    'has_constituent',
                    'enabled_by',
                ]
                or ('URIorCURIE' in rangelist and rangelist[-1] != 'IriType')
            ) and slotname != 'predicate':
                if slot.required:
                    validators.append(
                        self._gen_required_validator(slotname, is_multivalued=slot.multivalued)
                    )
                elif slot.multivalued:
                    validators.append(self._gen_scalar_to_list_check_curies_validator(slotname))
                else:
                    validators.append(self._gen_curie_prefix_validator(slotname))
            elif 'Entity' in rangelist:
                if len(rangelist) > 1:
                    if slot.required:
                        validators.append(
                            self._gen_required_validator(slotname, rangelist[-1], slot.multivalued)
                        )
                    elif slot.multivalued:
                        validators.append(
                            self._gen_scalar_to_list_check_curies_validator(slotname, rangelist[-1])
                        )
                    else:
                        validators.append(self._gen_curie_prefix_validator(slotname, rangelist[-1]))
                else:
                    if slot.required:
                        validators.append(
                            self._gen_required_validator(slotname, is_multivalued=slot.multivalued)
                        )
                    elif slot.multivalued:
                        validators.append(self._gen_scalar_to_list_check_curies_validator(slotname))
                    else:
                        validators.append(self._gen_curie_prefix_validator(slotname))
            elif slot.required:
                validators.append(
                    self._gen_required_validator(
                        slotname, is_multivalued=slot.multivalued, is_curie=False
                    )
                )
            elif slot.multivalued:
                validators.append(self._gen_scalar_to_list_check_curies_validator(slotname))

        if validators:
            ret_val = "\n\t# Validators\n\t" + "\n\t".join(validators)
        else:
            ret_val = ""

        return ret_val

    @staticmethod
    def _gen_curie_prefix_validator(slotname: str, cls_name: str = 'cls') -> str:
        """
        :param namespaces:
        :return:
        """
        return f'''
    @validator('{slotname}')
    def check_{slotname}_prefix(cls, field):
        check_curie_prefix({cls_name}, field)
        return field'''

    @staticmethod
    def _gen_scalar_to_list_check_curies_validator(slotname: str, cls_name: str = 'cls') -> str:
        """
        :param namespaces:
        :return:
        """
        return f'''
    @validator('{slotname}')
    def convert_{slotname}_to_list_check_curies(cls, field):
        return convert_scalar_to_list_check_curies({cls_name}, field)'''

    @staticmethod
    def _gen_required_validator(
        slotname: str, cls_name: str = 'cls', is_multivalued=False, is_curie=True
    ) -> str:
        """
        :param namespaces:
        :return:
        """
        validator = f'''
    @validator('{slotname}')
    def validate_required_{slotname}(cls, field):
        check_value_is_not_none("{slotname}", field)
'''
        if is_multivalued:
            validator += f'\t\tconvert_scalar_to_list_check_curies({cls_name}, field)\n'
        elif is_curie:
            validator += f'\t\tcheck_curie_prefix({cls_name}, field)\n'
        validator += f'\t\treturn field'
        return validator

    @staticmethod
    def _get_entity_post_init() -> str:
        """
        Post init for entity for inferring categories from the
        classes in its method resolution order

        requires a special classvar _category which is excluded
        from mixins
        """
        return '''
    def __post_init__(self):
        # Initialize default categories if not set
        # by traversing the MRO chain
        if not self.category:
            self.category = list(
                {
                    f'biolink:{super_class._category}'
                    for super_class in inspect.getmro(type(self))
                    if hasattr(super_class, '_category')
                }
            )
        '''

    @staticmethod
    def _get_pydantic_config_and_validators() -> str:
        """
        Pydantic config class and validator methods

        See https://pydantic-docs.helpmanual.io/usage/validators/#reuse-validators
        """
        return '''
class PydanticConfig:
    """
    Pydantic config
    https://pydantic-docs.helpmanual.io/usage/model_config/
    """

    validate_assignment = True
    validate_all = True
    underscore_attrs_are_private = True
    extra = 'forbid'
    arbitrary_types_allowed = True  # TODO re-evaluate this


def check_curie_prefix(cls, curie: Union[List, str, None]):
    if isinstance(curie, list):
        for cur in curie:
            prefix = cur.split(':')[0]
            if prefix not in valid_prefix:

                LOG.warning(f"{curie} prefix '{prefix}' not in curie map")
                if hasattr(cls, '_id_prefixes') and cls._id_prefixes:
                    LOG.warning(f"Consider one of {cls._id_prefixes}")
                else:
                    LOG.warning(
                        f"See https://biolink.github.io/biolink-model/context.jsonld "
                        f"for a list of valid curie prefixes"
                    )
    elif curie:
        prefix = curie.split(':')[0]
        if prefix not in valid_prefix:
            LOG.warning(f"{curie} prefix '{prefix}' not in curie map")
            if hasattr(cls, '_id_prefixes') and cls._id_prefixes:
                LOG.warning(f"Consider one of {cls._id_prefixes}")
            else:
                LOG.warning(
                    f"See https://biolink.github.io/biolink-model/context.jsonld "
                    f"for a list of valid curie prefixes"
                )          


def convert_scalar_to_list_check_curies(cls, field: Any) -> List[str]:
    """
    Converts list fields that have been passed a scalar to a 1-sized list
    
    Also checks prefix checks curies.  Because curie regex constraints
    are applied prior to running this function, we can use this for both
    curie and non-curie fields by rechecking re.match(curie_pattern, some_string)
    """
    if not isinstance(field, list):
        field = [field]
    for feld in field:
        if isinstance(feld, str) and re.match(curie_pattern, feld):
            check_curie_prefix(cls, feld)
    return field
    

def check_value_is_not_none(slotname: str, field: Any) -> bool:
    is_none = False
    if isinstance(field, list) or isinstance(field, dict):
        if not field:
            is_none = True
    else:
        if field is None:
            is_none = True
            
    if is_none:
        raise ValueError(f"{slotname} is required")
        '''


def main(yamlfile: str):
    pydantic_generator = PydanticGen(yamlfile)
    print(pydantic_generator.serialize())


if __name__ == "__main__":
    typer.run(main)



"""
Helper functions for facade files

Currently this is implemented with exceptions, but this
assumes that all "events" will trigger the code
to break out of its lexical scope

Some other options described here if this needs to be
more complex: https://stackoverflow.com/a/16192256
"""
from koza.exceptions import NextRowException


def next_row():
    """
    Breaks out of the facade file and iterates to the next row
    in the file

    Effectively a continue statement
    https://docs.python.org/3/reference/simple_stmts.html#continue
    :return:
    """
    raise NextRowException



from koza.manager.data_provider import inject_row, inject_map

source_name = "custom_entrez_2_string"

row = inject_row(source_name)

map = inject_map(source_name)

entry = dict()

entry["entrez"] = row["entrez"]

map[row["STRING"]] = entry



import pytest
from biolink_model_pydantic.model import Curie, Gene, GeneToGeneAssociation, Predicate, Publication

from koza.converter.kgx_converter import KGXConverter


def test_gene_conversion():
    fgf8a = Gene(
        id=Curie("ZFIN:ZDB-GENE-990415-72"),
        symbol="fgf8a",
        name="fibroblast growth factor 8a",
        in_taxon=["NCBITaxon:7955"],
    )

    kgx_converter: KGXConverter = KGXConverter()
    (nodes, edges) = kgx_converter.convert([fgf8a])
    output = nodes[0]

    assert output['id'] == "ZFIN:ZDB-GENE-990415-72"
    assert output['symbol'] == "fgf8a"
    assert 'gene' in str(output['category']).lower()
    assert output['name'] == "fibroblast growth factor 8a"
    assert output['in_taxon'] == ["NCBITaxon:7955"]


def test_association_conversion():
    fgf8a = Gene(
        id=Curie("ZFIN:ZDB-GENE-990415-72"), symbol="fgf8a", name="fibroblast growth factor 8a"
    )
    pax2a = Gene(id=Curie("ZFIN:ZDB-GENE-990415-8"), symbol="pax2a", name="paired box 2a")
    pub = Publication(id=Curie("PMID:17522161"), type="MESH:foobar")
    association = GeneToGeneAssociation(
        id='uuid:123',
        subject=fgf8a.id,
        predicate=Predicate.interacts_with,
        object=pax2a.id,
        publications=[pub],
        relation="RO:0003003",
    )

    (nodes, edges) = KGXConverter().convert([fgf8a, pax2a, pub, association])

    output = edges[0]
    assert output['subject'] == Curie("ZFIN:ZDB-GENE-990415-72")
    assert output['relation'] == Curie("RO:0003003")
    assert output['object'] == Curie("ZFIN:ZDB-GENE-990415-8")
    # TODO figure out how/where to handle this conversion
    # assert Curie("PMID:17522161") in output['publications']


@pytest.mark.parametrize(
    "id, symbol, synonym, xref",
    [
        ("MGI:1917258", "Ace2", ["RIKEN cDNA 2010305L05 gene", "2010305L05Rik"], []),
        ("RGD:620474", "Sox9", [], ["ENSEMBL:ENSRNOG00000002607", "NCBI_Gene:140586"]),
    ],
)
def test_keys_uniformity(id, symbol, synonym, xref):
    """
    Connfirm that the result of the conversion has all of the same fields, even if some aren't used
    """
    gene = Gene(id=Curie(id), symbol=symbol, synonym=synonym, xref=xref)

    (nodes, edges) = KGXConverter().convert([gene])
    output = nodes[0]

    assert 'category' in output.keys()
    assert 'id' in output.keys()
    assert 'symbol' in output.keys()
    assert 'synonym' in output.keys()
    assert 'xref' in output.keys()
    assert 'description' in output.keys()
    assert 'source' in output.keys()



import re
import uuid

from biolink_model_pydantic.model import Protein, PairwiseGeneToGeneInteraction, Predicate
from koza.manager.data_collector import write

from koza.koza_runner import get_koza_app
from koza.manager.data_provider import inject_translation_table

koza = get_koza_app()

source_name = 'protein-links-detailed'

translation_table = inject_translation_table()

for row in koza.file_registry[source_name]:

    protein_a = Protein(id='ENSEMBL:' + re.sub(r'\d+\.', '', row['protein1']))
    protein_b = Protein(id='ENSEMBL:' + re.sub(r'\d+\.', '', row['protein2']))

    pairwise_gene_to_gene_interaction = PairwiseGeneToGeneInteraction(
        id="uuid:" + str(uuid.uuid1()),
        subject=protein_a.id,
        object=protein_b.id,
        predicate=Predicate.interacts_with,
        relation = translation_table.global_table['interacts with'],
    )

    write(source_name, protein_a, protein_b, pairwise_gene_to_gene_interaction)



import pytest
from biolink_model_pydantic.model import (
    Gene,
    NamedThingToInformationContentEntityAssociation,
    Publication,
)

from koza.koza_runner import get_translation_table


@pytest.fixture
def gene_literature_entities(mock_koza):
    row = iter(
        [
            {
                'xb_article': 'XB-ART-1',
                'pmid': '16938438',
                'gene_pages': 'XB-GENEPAGE-487723 nog,XB-GENEPAGE-490377 msx1,XB-GENEPAGE-481828 bmp2,XB-GENEPAGE-483057 bmp4,XB-GENEPAGE-480982 fgf8,XB-GENEPAGE-6045068 hspa1l,XB-GENEPAGE-485625 hsp70',
            }
        ]
    )
    map_cache = {
        'genepage-2-gene': {
            'XB-GENEPAGE-487723': {
                'tropicalis_id': 'XB-GENE-487724',
                'laevis_l_id': 'XB-GENE-864891',
                'laevis_s_id': 'XB-GENE-17332089',
            },
            'XB-GENEPAGE-490377': {
                'tropicalis_id': 'XB-GENE-490378',
                'laevis_l_id': 'XB-GENE-490382',
                'laevis_s_id': 'XB-GENE-6253888',
            },
            'XB-GENEPAGE-481828': {
                'tropicalis_id': 'XB-GENE-481829',
                'laevis_l_id': 'XB-GENE-6252323',
                'laevis_s_id': 'XB-GENE-481833',
            },
            'XB-GENEPAGE-483057': {
                'tropicalis_id': 'XB-GENE-483058',
                'laevis_l_id': 'XB-GENE-483062',
                'laevis_s_id': 'XB-GENE-6254046',
            },
            'XB-GENEPAGE-480982': {
                'tropicalis_id': 'XB-GENE-480983',
                'laevis_l_id': 'XB-GENE-865429',
                'laevis_s_id': 'XB-GENE-17330198',
            },
            'XB-GENEPAGE-6045068': {
                'tropicalis_id': 'XB-GENE-6045069',
                'laevis_l_id': 'XB-GENE-17342842',
                'laevis_s_id': 'XB-GENE-6254949',
            },
            'XB-GENEPAGE-485625': {
                'tropicalis_id': 'XB-GENE-485626',
                'laevis_l_id': 'XB-GENE-485630',
                'laevis_s_id': 'XB-GENE-17340402',
            },
        }
    }
    tt = get_translation_table("examples/translation_table.yaml", None)

    return mock_koza(
        'gene-literature',
        row,
        './examples/xenbase/gene-literature.py',
        map_cache=map_cache,
        translation_table=tt,
    )


def test_gene_literature_entities(gene_literature_entities):
    assert gene_literature_entities


def test_gene_literature_entity_types(gene_literature_entities):
    publications = [
        entity for entity in gene_literature_entities if isinstance(entity, Publication)
    ]
    genes = [entity for entity in gene_literature_entities if isinstance(entity, Gene)]
    associations = [
        entity
        for entity in gene_literature_entities
        if isinstance(entity, NamedThingToInformationContentEntityAssociation)
    ]

    # a roundabout way of confirming that everything generated is one of these three and there's nothing else
    assert len(gene_literature_entities) == len(publications) + len(genes) + len(associations)
    assert len(publications) == 1


def test_gene_literature_entity(gene_literature_entities):
    genes = [entity for entity in gene_literature_entities if isinstance(entity, Gene)]

    # confirm one gene id from each species
    assert 'Xenbase:XB-GENE-480983' in [gene.id for gene in genes]
    assert 'Xenbase:XB-GENE-6253888' in [gene.id for gene in genes]
    assert 'Xenbase:XB-GENE-485630' in [gene.id for gene in genes]



"""
Testing the biolink config dataclasses + pydandic

"""
from pathlib import Path

import pytest
import yaml

from koza.model.config.source_config import PrimaryFileConfig

base_config = Path(__file__).parent / 'resources' / 'primary-source.yaml'


def test_source_primary_config():
    with open(base_config, 'r') as config:
        PrimaryFileConfig(**yaml.safe_load(config))


@pytest.mark.parametrize(
    "inclusion, column, filter_code, value",
    (
        [
            ('include', 'combined_score', 'lt', '70'),
            ('exclude', 'combined_score', 'lt', '70'),
            ('include', 'combined_score', 'in', '70'),
            ('exclude', 'combined_score', 'in', '70'),
            ('exclude', 'combined_score', 'in', 70),
            ('exclude', 'combined_score', 'in', 0.7),
            ('include', 'combined_score', 'eq', ['goat', 'sheep']),
            ('include', 'combined_score', 'lt', ['goat', 'sheep']),
            ('include', 'combined_score', 'gte', ['goat', 'sheep']),
            ('exclude', 'is_ungulate', 'eq', 'T'),
        ]
    ),
)
def test_wrong_filter_type_raises_exception(inclusion, column, filter_code, value):
    """
    Test if include and exclude raise a
    value error when handed an incompatible type,
    eg a string when using the lt operator
    """
    with open(base_config, 'r') as config:
        source_config = yaml.safe_load(config)
        del source_config['filters']

        source_config['filters'] = [
            {'column': column, 'inclusion': inclusion, 'filter_code': filter_code, 'value': value}
        ]
        with pytest.raises(ValueError):
            PrimaryFileConfig(**source_config)


@pytest.mark.parametrize("inclusion, code", [('include', 'lgt'), ('exclude', 'ngte')])
def test_wrong_filter_code_raises_exception(inclusion, code):
    with open(base_config, 'r') as config:
        source_config = yaml.safe_load(config)
        source_config['filters'] = [
            {'column': 'combined_score', 'inclusion': inclusion, 'filter_code': code, 'value': 70}
        ]
        with pytest.raises(ValueError):
            PrimaryFileConfig(**source_config)
